from datetime import datetime
import pytz
import threading

import discord

from src.services.notification_type import NotificationType
from src.services.server_service import ServerService


class NotificationService:
    def __init__(self, config, server_service: ServerService):
        self.config = config
        self.servers = server_service

        self.lock_emojis = {
            # Keys are the boolean int values returned by "password_protected"
            1: discord.PartialEmoji(name=":lock:"),
            0: discord.PartialEmoji(name=":unlock:")
        }

        self.should_ping_role = True
        self.timeout_timer = threading.Timer(float(self.config["ping_timeout"]), self.restart_role_ping)

    def fetch_populated_servers_info(self):
        servers = self.servers.get_server_list()
        return self.servers.get_populated_servers(servers)

    def get_notification_type(self, last_msg_author_id, bot_id, servers, is_update_forced) -> NotificationType:
        """Contains all the logic involved in determining characteristics of notification message.
           Will set a ping timeout timer if determined that next notification should ping.
        """
        if len(servers) == 0:
            if is_update_forced:
                return NotificationType.Forced

            return NotificationType.NoNotification

        # If the message at the bottom of chat is bot's update then edit it with new values to reduce spamming
        if (last_msg_author_id == bot_id) is False or self.should_ping_role:
            # Editing messages does not trigger pings, so send a message if ping timeout passed
            if self.should_ping_role:
                self.timeout_role_ping()

            return NotificationType.New

        return NotificationType.Edit

    def create_msg_content(self, populated_servers, role_mention):
        if len(populated_servers) == 0:
            return "All servers are empty."

        update_time = datetime.now()
        timezone = pytz.timezone("CET")
        update_time_localized = timezone.localize(update_time)
        message = f'STATUS FOR: {update_time_localized.strftime("%H:%M")} CET\n\n'

        if self.should_ping_role and \
                any(server["player_count"] >= self.config["min_ping_players"] for server in populated_servers):
            message += f"{role_mention}\n People are playing on:\n\n"
        else:
            message += "People are playing on:\n\n"

        for server in populated_servers:
            server_protected_emoji = self.lock_emojis[server["password_protected"]]
            message += f'{server["server_name"]} {server_protected_emoji} - {server["map"]} \n'
            message += f'Players: {server["player_count"]} / {server["max_players"]} \n\n'

        return message

    def timeout_role_ping(self):
        if self.should_ping_role is True:
            self.should_ping_role = False

            # Timer objects need to be created for each timeout period
            # https://docs.python.org/3/library/threading.html#threading.Thread.start
            self.timeout_timer = threading.Timer(float(self.config["ping_timeout"]), self.restart_role_ping)
            self.timeout_timer.daemon = True
            self.timeout_timer.start()

    def restart_role_ping(self):
        self.should_ping_role = True
