import re
from concurrent.futures.thread import ThreadPoolExecutor
from io import StringIO

import valve.source
import valve.source.a2s
import valve.source.master_server
from markdown import Markdown

class ServerService:
    def __init__(self, config):
        self.config = config

    def get_server_list(self):
        with valve.source.master_server.MasterServerQuerier() as msq:
            return [address for address in msq.find(gamedir=u"NeotokyoSource")]

    def get_populated_servers(self, server_list):
        with ThreadPoolExecutor() as ex:
            futures = [ex.submit(self.get_populated_server_info, address) for address in
                       server_list]

        results = [f.result() for f in futures]
        # Servers that got returned as None are not pubbing
        return [server for server in results if server is not None]

    def get_populated_server_info(self, address):
        try:
            with valve.source.a2s.ServerQuerier(address) as server:
                info = server.info()
                players = server.players()

                # Reduce amount of object juggling by appending all info to one object
                info["player_count"] = players["player_count"]
                info["server_name"] = self._sanitize_name(info["server_name"])

                if self._is_server_pubbing(info):
                    return info

        except valve.source.a2s.NoResponseError:
            return None

    def _is_server_pubbing(self, info) -> bool:
        # Locked servers are not pubbing, so skip them
        if info["password_protected"] == 1:
            return False

        # AFAIK bot is not included in "player_count" so no need to readjust it manually
        if info["player_count"] < self.config["min_players"]:
            return False

        return True

    def _sanitize_name(self, server_name):
        name = self._remove_links_from_name(server_name)
        name = self._remove_mentions_from_name(name)
        name = self._remove_markdown_from_name(name)

        return name

    def _remove_links_from_name(self, server_name):
        # This pattern works for most URL I tested, but not for IP:Port
        pattern = r'(?i)\b((?:[a-z][\w-]+:(?:/{1,3}|[a-z0-9%])|www\d{0,3}[.]|[a-z0-9.\-]+[.][a-z]{2,4}/)(?:[^\s()<>]+|\(([^\s()<>]+|(\([^\s()<>]+\)))*\))+(?:\(([^\s()<>]+|(\([^\s()<>]+\)))*\)|[^\s`!()\[\]{};:\'".,<>?«»“”‘’]))';

        ip_port_pattern = r'\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}:\d{1,5}'

        safe_name = re.sub(pattern, "[LINK REMOVED]", server_name)
        safe_name = re.sub(ip_port_pattern, "[LINK REMOVED]", safe_name)

        return safe_name

    def _remove_mentions_from_name(self, server_name):
        safe_name = re.sub("@", "", server_name)

        return safe_name

    def _remove_markdown_from_name(self, server_name):
        def unmark_element(element, stream=None):
            if stream is None:
                stream = StringIO()
            if element.text:
                stream.write(element.text)
            for sub in element:
                unmark_element(sub, stream)
            if element.tail:
                stream.write(element.tail)
            return stream.getvalue()

        # patching Markdown
        Markdown.output_formats["plain"] = unmark_element
        __md = Markdown(output_format="plain")
        __md.stripTopLevelTags = False

        return __md.convert(server_name)
